from typing import Any

import aiomqtt

from ..messages import (
    Message,
)
from . import MessageEncoder, MqttClient, MqttConnectionError, Will


class PahoMqttClient(MqttClient):
    def __init__(
        self,
        config: dict[str, Any],
        encoder: MessageEncoder,
    ):
        self._config = config
        self._client = None
        self._client_id = None
        self._encoder = encoder

    async def connect(self, component_name: str, will: Will):
        payload = self._encoder.encode(will.message.get_payload())
        will_ = aiomqtt.Will(
            will.message.topic, payload, qos=will.qos, retain=will.retain
        )
        self._client_id = component_name

        tls_pars = None
        if self._config.get("ca_certs"):
            tls_pars = aiomqtt.TLSParameters(ca_certs=self._config["ca_certs"])

        self._client = aiomqtt.Client(
            self._config["hostname"],
            int(self._config["port"]),
            identifier=component_name,
            username=self._config["username"],
            password=self._config["password"],
            will=will_,
            tls_params=tls_pars,
            keepalive=30,
        )

        await self._client.__aenter__()  # Connect with aiomqtt

    async def publish(self, message: Message, qos: int, retain: bool):
        if not self._client or not self._client_id:
            raise MqttConnectionError("Client not connected to Mqtt Server")

        payload = self._encoder.encode(message.get_payload())
        await self._client.publish(message.topic, payload, qos=qos, retain=retain)

    async def subscribe(self, topic: str, qos: int):
        if not self._client or not self._client_id:
            raise MqttConnectionError("Client not connected to Mqtt Server")

        await self._client.subscribe(topic, qos=qos)

    async def deliver_message(self) -> Message:
        # if self._client is None:
        #     raise MqttConnectionError(f"Client is not connected, so no messages")

        message = await anext(self._client.messages)

        payload = self._encoder.decode(message.payload.decode())
        payload["topic"] = message.topic.value
        return Message.from_payload(payload)

    async def disconnect(self):
        if self._client is None:
            return

        await self._client.__aexit__(None, None, None)



"""Main components of sparkplug standard"""
import abc
import asyncio
import logging
from functools import singledispatchmethod
from queue import Queue
from typing import Literal

from aiospb.mqtt import MqttClient, Will

from . import Clock, RealClock
from . import messages as m
from .data import DataType, Metric, MetricChange, MetricChangeRequest, ValueType

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

_NODES_STATES = Literal["online", "offline", "publishing"]


class DeviceConnectionError(Exception):
    ...


class DeviceDriver(abc.ABC):
    """Connection to real devices or direct interfaces"""

    @abc.abstractmethod
    async def load_metrics(self) -> list[Metric]:
        """Load metrics from system"""

    @abc.abstractmethod
    async def scan_metrics(self, aliases: list[int] | None = None) -> list[Metric]:
        """Scan the device for last state of metrics"""

    @abc.abstractmethod
    async def write_metric(self, request: MetricChangeRequest) -> MetricChange:
        """Write metric to system"""


class HistoricalStore(abc.ABC):
    """Store historical changes to retrieve when host application is online"""

    @abc.abstractmethod
    async def save(self, changes: list[MetricChange]):
        """Save all change metrics of a scan"""

    @abc.abstractmethod
    async def load(self) -> list[MetricChange]:
        """Load all historical metric changes"""

    @abc.abstractmethod
    async def has_data(self) -> bool:
        """Has data stored?"""

    @abc.abstractmethod
    async def clear(self):
        """Clear content of historical"""


class OnmemHistoricalStore(HistoricalStore):
    """Store historical changes on memory. Not recomemded if host will be stopped for a long time"""

    def __init__(self):
        self._changes = []

    async def save(self, changes: list[MetricChange]):
        self._changes.extend(changes)

    async def load(self) -> list[MetricChange]:
        changes = [
            MetricChange(
                change.ts,
                change.value,
                change.data_type,
                change.alias,
                change.metric_name,
                change.properties,
                is_historical=True,
            )
            for change in self._changes
        ]
        self._changes.clear()
        return changes

    async def has_data(self):
        return bool(self._changes)

    async def clear(self):
        self._changes.clear()


class EdgeNode:
    """Gateway connected to mqtt and to one or more hardware devices"""

    def __init__(
        self,
        name: str,
        group_name: str,
        mqtt_client: MqttClient,
        device_driver: DeviceDriver,
        historical_store: HistoricalStore | None = None,
        primary_hostname: str = "",
        scan_rate: int = 60000,  # Scan every minute
        clock: Clock | None = None,
    ):
        self._name = f"{group_name}/{name}"
        self._primary_hostname = primary_hostname
        self._clock = clock if clock else RealClock()

        self._driver = device_driver
        self._client = mqtt_client
        self._historical_store = (
            historical_store if historical_store else OnmemHistoricalStore()
        )

        self._seq = 255
        self._inner_metrics = {
            "bdSeq": m.Metric(
                "bdSeq", self._clock.now(), 255, data_type=DataType.Int64
            ),
            "Node Control/Rebirth": m.Metric(
                "Node Control/Rebirth",
                self._clock.now(),
                False,
                data_type=DataType.Boolean,
            ),
            "Node Control/Reboot": m.Metric(
                "Node Control/Reboot",
                self._clock.now(),
                False,
                data_type=DataType.Boolean,
            ),
            "Node Control/Scan Rate": m.Metric(
                "Node Control/Scan Rate",
                self._clock.now(),
                scan_rate,
                data_type=DataType.Int64,
            ),
        }
        self._metrics_by_name: dict[str, Metric] = {}
        self._metrics_by_alias: dict[int, Metric] = {}

        self._incomings_task = None
        self._outcomings_task = None
        self._scanning_task = None
        self._changes_queue = Queue()
        self._state: _NODES_STATES = "offline"

    @property
    def name(self) -> str:
        """Return name of edge of network node"""
        return self._name.split("/")[1]

    @property
    def group_name(self) -> str:
        """Return the name of the group of nodes it beyong to"""
        return self._name.split("/")[0]

    @property
    def state(self) -> _NODES_STATES:
        """Return the current state of node"""
        return self._state

    @property
    def primary_hostname(self) -> str:
        """Name of primary host application"""
        return self._primary_hostname

    @property
    def metrics(self) -> dict[str, Metric]:
        """Return a dict with all the metrics of the node"""
        metrics = self._inner_metrics.copy()
        metrics.update(self._metrics_by_name)
        return metrics

    def clear_metrics(self):
        """Remove all not internal metrics"""
        self._metrics_by_name.clear()
        self._metrics_by_alias.clear()

    def _add_metric(self, metric: Metric):
        """Add new metric in node"""
        self._metrics_by_name[metric.name] = metric
        if metric.alias:
            self._metrics_by_alias[metric.alias] = metric

    def _get_seq(self):
        self._seq = self._seq + 1 if self._seq != 255 else 0
        return self._seq

    def _get_bd_seq(self):
        bd_seq = self._inner_metrics["bdSeq"]
        value = int(bd_seq.value) + 1 if bd_seq.value != 255 else 0
        bd_seq = Metric("bdSeq", self._clock.now(), value, data_type=DataType.Int64)
        self._inner_metrics["bdSeq"] = bd_seq
        return value

    async def _publish_data_contents(self):
        while True:
            if self._state == "publishing":
                # Publish Ndata messages
                changes = []
                while not self._changes_queue.empty():
                    changes.append(self._changes_queue.get())

                if not changes:
                    await self._clock.asleep(0)
                    continue

                logger.info(f"Publishing {len(changes)} changes from node {self.name}")
                await self._client.publish(
                    m.Message.create(
                        self._name,
                        m.NdataContent(self._get_seq(), changes),
                        self._clock,
                    ),
                    qos=0,
                    retain=False,
                )
            else:
                await self._clock.asleep(0)

    async def _scan_metrics(self):
        scan_rate = int(self._inner_metrics["Node Control/Scan Rate"].value)
        while True and scan_rate:
            now = self._clock.now()
            next_scan_time = (now // scan_rate + 1) * scan_rate
            await self._clock.asleep((next_scan_time - now) / 1000)

            metrics = await self._driver.scan_metrics()

            changes = []
            for metric in metrics:
                old_metric = self._metrics_by_name[metric.name]
                change = old_metric.compare(metric)
                if change:
                    self._add_metric(metric)
                    changes.append(change)

            logger.debug(
                f"{self._name} has detected {len(changes)} changes to be published"
            )

            if self._state == "publishing":
                if await self._historical_store.has_data():
                    logger.debug(f"Getting historical data from node {self.name}")
                    for change in await self._historical_store.load():
                        self._changes_queue.put(change)
                    await self._historical_store.clear()

                for change in changes:
                    self._changes_queue.put(change)
            else:
                logger.debug(f"{self._name} is storing changes in historical")
                await self._historical_store.save(changes)

    async def _handle_incomming_messages(self):
        while True:
            message: m.Message = await self._client.deliver_message()
            await self._handle_content(message.content)

    @singledispatchmethod
    async def _handle_content(self, content: m.MessageContent):
        raise NotImplementedError(
            f"Content type {type(content)} is not managed by a Edge Node"
        )

    @_handle_content.register
    async def _handle_state_content(self, content: m.StateContent):
        if content.online:
            self._state = "publishing"
            await self._publish_birth_certificate()
        else:
            self._state = "online"

    @_handle_content.register
    async def _handle_command_content(self, content: m.NcmdContent):
        logger.debug(f"{self._name} has recieved command {content}")
        for request in content.requests:
            if request.metric_name == "Node Control/Rebirth" and request.value is True:
                self._state = "online"  # Stops sending data
                # topic = f"spBv1.0/{self.group_name}/NDATA/{self.name}"
                await self._client.publish(
                    m.Message.create(
                        self._name,
                        m.NdataContent(
                            self._get_seq(),
                            [
                                MetricChange(
                                    self._clock.now(),
                                    True,
                                    DataType.Boolean,
                                    metric_name="Node Control/Rebirth",
                                )
                            ],
                        ),
                        self._clock,
                    ),
                    qos=0,
                    retain=False,
                )
                await self._publish_birth_certificate()
            if request.metric_name == "Node Control/Reboot" and request.value is True:
                await self.terminate_session()
                await self.establish_session()
            else:
                loop = asyncio.get_event_loop()
                loop.create_task(self._process_metric_change(request))

    async def _process_metric_change(self, request: MetricChangeRequest):
        change = await self._driver.write_metric(request)
        logger.debug(f"{self._name} has procesed change {change}")

        metric = (
            self._metrics_by_alias[change.alias]
            if change.alias
            else self._metrics_by_name[change.metric_name]
        )

        self._add_metric(
            Metric(
                metric.name,
                change.timestamp,
                change.value,
                metric.data_type,
                metric.properties,
                metric.alias,
                metric.is_transient,
            )
        )
        self._changes_queue.put(change)

    async def _publish_birth_certificate(self):
        self.clear_metrics()
        metrics = await self._driver.load_metrics()

        for metric in metrics:
            self._add_metric(metric)

        logger.debug(
            f"{self._name} publishes birth certificate with {len(metrics)} metrics"
        )
        await self._client.publish(
            m.Message.create(
                self._name,
                m.NbirthContent(self._get_seq(), list(self.metrics.values())),
                self._clock,
            ),
            qos=0,
            retain=False,
        )
        self._state = "publishing"

    async def establish_session(self):
        """Create a session"""

        self._get_bd_seq()
        await self._client.connect(
            self._name,
            will=Will(
                m.Message.create(
                    self._name,
                    m.NdeathContent(self._inner_metrics["bdSeq"]),
                    self._clock,
                ),
                0,
                False,
            ),
        )
        await self._client.subscribe("spBv1.0/STATE/+", qos=1)
        await self._client.subscribe(
            f"spBv1.0/{self.group_name}/NCMD/{self.name}", qos=0
        )

        self._state = "online"

        self._outcomings_task = asyncio.create_task(self._publish_data_contents())
        self._incomings_task = asyncio.create_task(self._handle_incomming_messages())
        self._scanning_task = asyncio.create_task(self._scan_metrics())

    async def terminate_session(self):
        """Finish a session cleanly, leaving the node clean"""
        if self._outcomings_task is not None:
            self._outcomings_task.cancel()
            self._outcomings_task = None
        if self._scanning_task is not None:
            self._scanning_task.cancel()
            self._scanning_task = None
        if self._incomings_task is not None:
            self._incomings_task.cancel()
            self._incomings_task = None

        self._state = "offline"
        if self._client:
            logger.info(f"Publishing clean Death Certificate from node {self.name}")
            await self._client.publish(
                m.Message.create(
                    self._name,
                    m.NdeathContent(self._inner_metrics["bdSeq"]),
                    self._clock,
                ),
                0,
                False,
            )

        await self._client.disconnect()



import abc
import asyncio
import logging
import re

from aiospb.mqtt import MqttClient
from aiospb.mqtt.encoding import JsonEncoder
from aiospb.mqtt.paho import PahoMqttClient
from aiospb.nodes import DeviceDriver, EdgeNode

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class MqttServer:
    """Create mqtt clients working asyncronously"""

    def __init__(self, mqtt_config: dict[str, str]):
        self._config = mqtt_config
        self._encoder = JsonEncoder()

    def create_client(self) -> MqttClient:
        """Create an asyncronous mqtt client"""
        return PahoMqttClient(self._config, self._encoder)


class NodeFactory(abc.ABC):
    """Create connections to device from Nodes/Devices"""

    @abc.abstractproperty
    def node_names(self) -> list[str]:
        """Return all name of nodes available to connect by"""

    @abc.abstractmethod
    def create(self, name: str) -> EdgeNode:
        """Create a device connection"""


class NodesGroup:
    """Runnable class which manage a group of edge nodes"""

    def __init__(
        self,
        name: str,
        node_fry: NodeFactory,
        time_to_restart: int = 10,
        cicle_time: int = 2,
    ):
        self._name = name
        self._node_fry = node_fry
        self._filter = None
        self._nodes = []
        self._time_to_restart = time_to_restart
        self._cicle_time = cicle_time

    @property
    def name(self) -> str:
        """Return the name of the group"""
        return self._name

    @property
    def nodes(self) -> list[EdgeNode]:
        """Return all executable nodes."""
        return self._nodes.copy()

    def setup(self, nodes_filter: str = ""):
        pattern = re.compile(nodes_filter)
        for name in self._node_fry.node_names:
            if pattern.match(name):
                self._nodes.append(self._node_fry.create(name))

    async def _start_all_nodes(self, cycles: int = -1):
        for node in self._nodes:
            logger.info(f"Establishing session at node {node.name}")
            await node.establish_session()

        counter = 0
        while True:
            if cycles >= 0:
                counter += 1
            for node in self._nodes:
                if node.state == "crashed":
                    logger.warning(
                        f"Node '{node.name}' has crashed internally, terminating cleanly"
                    )
                    await node.terminate_session()
                    await asyncio.sleep(self._time_to_restart)

                    logger.info(f"Restablishing session at node {node.name}")
                    await node.establish_session()
            if cycles and counter == cycles:
                break
            await asyncio.sleep(self._cicle_time)

    def run(self, cycles: int = -1):
        """Run the group as an application in the OS"""
        asyncio.run(self._start_all_nodes(cycles))



import json
from typing import Any

from .. import messages as m
from . import MessageEncoder


class JsonEncoder(MessageEncoder):
    """Encoder of messages for testing purposes"""

    def encode(self, payload: dict[str, Any]) -> str:
        """Convert a payload dict to a string for publishing"""

        return json.dumps(payload, sort_keys=True)

    def decode(self, payload: str) -> dict[str, Any]:
        """Convert payload to a payload dict"""
        return json.loads(payload)



import abc
from dataclasses import dataclass
from typing import Any

from aiospb.messages import Message


class MqttConnectionError(Exception):
    ...


@dataclass
class Will:
    message: Message
    qos: int
    retain: bool


class MessageEncoder(abc.ABC):
    """Encode the message before sending it as payload in the message"""

    @abc.abstractmethod
    def encode(self, payload_dict: dict[str, Any]) -> str:
        """Convert a message to a payload"""

    @abc.abstractmethod
    def decode(self, payload: str) -> dict[str, Any]:
        """Convert payload to a message object"""


class MqttClient(abc.ABC):
    @abc.abstractmethod
    async def connect(self, component_name: str, will=Will):
        """Connect a component to MQTT server"""

    @abc.abstractmethod
    async def publish(self, topic: str, message: Message, qos: int, retain: bool):
        """Publish a message  to the topic"""

    @abc.abstractmethod
    async def deliver_message(self) -> Message:
        """Return a messsage recieved from the MQTT Server"""

    @abc.abstractmethod
    async def subscribe(self, topic: str, qos: int):
        """Subscribe the component to recieve messages from a topic"""

    @abc.abstractmethod
    async def disconnect(self):
        """Disconnect the client from the MQTT server"""



import abc
import asyncio
import datetime
import time


class Clock(abc.ABC):
    @abc.abstractmethod
    def now(self) -> int:
        """Return current timestamp in ms"""

    @abc.abstractmethod
    def sleep(self, seconds: float):
        """Syncronous sleep for several seconds"""

    @abc.abstractmethod
    async def asleep(self, seconds: float):
        """Asyncronous sleep for several seconds"""


class RealClock(Clock):
    def now(self) -> int:
        return int(datetime.datetime.now(datetime.timezone.utc).timestamp() * 1000)

    def sleep(self, seconds: float):
        time.sleep(seconds)

    async def asleep(self, seconds: float):
        await asyncio.sleep(seconds)



import asyncio
from typing import Callable, Coroutine
import logging

from aiospb import Clock, RealClock
from aiospb.data import Metric, MetricChange, MetricChangeRequest, ValueType
from aiospb.groups import MqttServer
from aiospb.messages import (
    Message,
    NcmdContent,
    StateContent,
    WarningContent,
)
from aiospb.mqtt import Will

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class HostApplication:
    """Implementation of Host Application"""

    def __init__(
        self,
        hostname: str,
        mqtt_server: MqttServer,
        *node_groups: str,
        clock: Clock | None = None,
        max_delay: int = 10,
    ):
        if not node_groups:
            raise ValueError("Not node groups to listen are defined")
        self._groups = node_groups

        self._client = mqtt_server.create_client()
        self._hostname = hostname
        self._callbacks = {}
        self._nodes = {}
        self._clock = clock if clock else RealClock()
        self._state = "offline"
        self._listen_nodes = None
        self._max_delay = max_delay
        self._node_sequences = {}

    @property
    def hostname(self) -> str:
        """Name of the host application"""
        return self._hostname

    @property
    def state(self):
        return self._state

    async def establish_session(self):
        """Init session to listen edge nodes"""

        await self._client.connect(
            self._hostname,
            will=Will(
                Message.create(self._hostname, StateContent(False), self._clock),
                1,
                True,
            ),
        )
        if self._groups:
            for group in self._groups:
                await self._client.subscribe(f"spBv1.0/{group}/#", qos=0)
        else:
            await self._client.subscribe("spBv1.0/+/+/+/#", qos=0)

        await self._client.publish(
            Message.create(self._hostname, StateContent(True), self._clock),
            qos=1,
            retain=True,
        )
        logger.debug(f'Host application "{self._hostname}" has established session')
        self._listen_nodes = asyncio.create_task(self._recieve_node_messages())

    def done(self):
        return self._listen_nodes is None or self._listen_nodes.done()

    def listen_nodes(
        self,
        callback: Callable[[Message], Coroutine[None, None, None]],
        node_filter: str | None = None,
    ) -> None:
        """Add one callable observer when it rece"""
        if node_filter not in self._callbacks:
            self._callbacks[node_filter] = []

        self._callbacks[node_filter].append(callback)

    async def _recieve_node_messages(self):
        while True:
            message: Message = await self._client.deliver_message()
            delay = (self._clock.now() - message.timestamp) / 1000
            await self._notify(message)
            logger.debug(
                f"Host app '{self._hostname}' has recieved {message.content_type} with delay of {delay} s"
            )

            if delay > self._max_delay:
                await self._notify(
                    Message.create(
                        message.component_name,
                        WarningContent(
                            "ExcessOfDelay",
                            f"Delay (seconds) is over max {self._max_delay} s",
                            [delay],
                        ),
                        self._clock,
                    )
                )
                logger.debug(
                    f'Host app "{self._hostname}" has notified warning by delay'
                )

    async def _notify(self, message):
        awaitables = [callback(message) for callback in self._callbacks[None]]
        node_name = message.component_name
        if node_name in self._callbacks:
            awaitables.extend(
                [callback(message) for callback in self._callbacks[node_name]]
            )
        await asyncio.gather(*awaitables)

    async def request_change_to_metrics(
        self, node_name: str, *change_requests: MetricChangeRequest
    ):
        """Request changes to metrics"""
        await self._client.publish(
            Message.create(node_name, NcmdContent(list(change_requests)), self._clock),
            qos=0,
            retain=False,
        )

    async def terminate_session(self):
        """Close cleanly a session"""
        if self._listen_nodes:
            self._listen_nodes.cancel()

        await self._client.publish(
            Message.create(self._hostname, StateContent(False), self._clock),
            qos=1,
            retain=True,
        )
        await self._client.disconnect()
        logger.debug(
            f'Host application "{self._hostname}" has cleanly terminated session'
        )



from dataclasses import dataclass, field, asdict
from enum import Enum
from typing import Any, Mapping, Self


class DataType(Enum):
    Unknown = 0
    Int8 = 1
    Int16 = 2
    Int32 = 3
    Int64 = 4
    UInt8 = 5
    UInt16 = 6
    UInt32 = 7
    UInt64 = 8
    Float = 9
    Double = 10
    Boolean = 11
    String = 12
    DateTime = 13
    Text = 14
    UUID = 15
    DataSet = 16
    Bytes = 17
    File = 18
    Template = 19
    PropertySet = 20
    PropertySetList = 21
    Int8Array = 22
    Int16Array = 23
    Int32Array = 24
    Int64Array = 25
    UInt8Array = 26
    UInt16Array = 27
    UInt32Array = 28
    UInt64Array = 29
    FloatArray = 30
    DoubleArray = 31
    BooleanArray = 32
    StringArray = 33
    DateTimeArray = 34


ValueType = bool | int | float | str | None


@dataclass(frozen=True)
class PropertyValue:
    data_type: DataType
    value: ValueType

    @classmethod
    def from_dict(cls, payload):
        return cls(
            DataType[payload["type"]],
            None if payload.get("is_null") else payload["value"],
        )

    def to_dict(self):
        result: dict[str, ValueType] = {"type": self.data_type.name}
        if self.value is None:
            result["is_null"] = True
        else:
            result["value"] = self.value
        return result


class QualityOptions(Enum):
    BAD = 0
    GOOD = 192
    STALE = 500


@dataclass(frozen=True)
class QualityValue(PropertyValue):
    """Standard property of metric"""

    def __init__(self, value: QualityOptions):
        if value not in QualityOptions:
            raise ValueError(f"The value {value} is not valid")
        super().__init__(DataType.Int32, value.value)


@dataclass
class MetricChange:
    timestamp: int
    value: int | bool | str | float | None
    data_type: DataType
    alias: int | None = None
    metric_name: str = ""
    properties: dict[str, PropertyValue] = field(default_factory=dict)
    is_historical: bool = False

    @classmethod
    def from_dict(cls, dict_value: dict[str, Any]) -> Self:
        """Create metric from dict definition"""

        properties = {
            key: PropertyValue.from_dict(value)
            for key, value in dict_value.get("properties", {}).items()
        }
        alias = dict_value.get("alias")
        metric_name = "" if alias is not None else dict_value.get("name", "")
        value = None if dict_value.get("is_null") else dict_value["value"]
        is_historical = True if dict_value.get("is_historical") else False

        return cls(
            dict_value["timestamp"],
            value,
            DataType[dict_value["dataType"]],
            alias,
            metric_name,
            properties,
            is_historical,
        )

    def to_dict(self) -> dict[str, ValueType]:
        """Create dict from metric change object"""

        result = {"timestamp": self.timestamp, "dataType": self.data_type.name}
        if self.alias:
            result["alias"] = self.alias
        else:
            result["name"] = self.metric_name
        if self.value is None:
            result["is_null"] = True
        else:
            result["value"] = self.value
        properties = {key: value.to_dict() for key, value in self.properties.items()}
        if properties:
            result["properties"] = properties

        if self.is_historical:
            result["is_historical"] = True

        return result

    # def update_metric(self, metric: Metric) -> Metric:
    #     """Update metric values from the change"""

    #     if self.timestamp <= metric.timestamp or self.is_historical:
    #         # Metric can not be updated with a previous change
    #         return metric

    #     if not self.properties and self.value == metric.value:
    #         return metric

    #     properties = {}
    #     for key, value in metric.properties.items():
    #         properties[key] = (
    #             value if key not in self.properties else self.properties[key]
    #         )

    #     return Metric(
    #         metric.name,
    #         self.timestamp,
    #         self.value,
    #         metric.data_type,
    #         properties,
    #         metric.alias,
    #         metric.is_transient,
    #     )


@dataclass(frozen=True)
class Metric:
    """Metric owned by a edge node or device"""

    name: str
    timestamp: int
    value: int | bool | str | float | None
    data_type: DataType
    properties: dict[str, PropertyValue] = field(default_factory=dict)
    alias: int | None = None
    is_transient: bool = False

    @classmethod
    def from_dict(cls, value: dict[str, Any]):
        properties = {
            key: PropertyValue.from_dict(value_)
            for key, value_ in value.get("properties", {}).items()
        }

        metric_value = None if value.get("is_null") else value.get("value")
        return cls(
            value.get("name", ""),
            value["timestamp"],
            metric_value,
            DataType[value["dataType"]],
            properties,
            value.get("alias", None),
            value.get("is_transient", False),
        )

    def to_dict(self):
        result = {
            "name": self.name,
            "timestamp": self.timestamp,
            "dataType": self.data_type.name,
        }
        if self.value is None:
            result["is_null"] = True
        else:
            result["value"] = self.value

        if self.is_transient:
            result["is_transient"] = True

        if self.alias:
            result["alias"] = self.alias

        properties = {}
        for key, value in self.properties.items():
            properties[key] = value.to_dict()

        if properties:
            result["properties"] = properties

        return result

    def compare(self, metric: Self) -> MetricChange | None:
        if metric.name != self.name:
            raise ValueError("Metrics has diferent names!!")

        if set(metric.properties.keys()) != set(self.properties.keys()):
            raise ValueError("Properties definitions has changed")

        if metric.timestamp <= self.timestamp:
            return

        if metric.value == self.value and metric.properties == self.properties:
            return

        properties = {}
        for key, value in metric.properties.items():
            if self.properties[key] != value:
                properties[key] = value

        alias = self.alias
        name = "" if alias else self.name
        return MetricChange(
            metric.timestamp, metric.value, self.data_type, alias, name, properties
        )


@dataclass(frozen=True)
class MetricChangeRequest:
    timestamp: int
    value: ValueType
    data_type: DataType
    metric_name: str = ""
    alias: int | None = None

    def to_dict(self):
        result = {
            "timestamp": self.timestamp,
            "value": self.value,
            "dataType": self.data_type.name,
        }
        if self.alias is not None:
            result["alias"] = self.alias
        else:
            result["name"] = self.metric_name
        return result

    @classmethod
    def from_dict(cls, payload: dict[str, Any]) -> Self:
        return cls(
            payload["timestamp"],
            payload["value"],
            DataType[payload["dataType"]],
            payload.get("name", ""),
            payload.get("alias"),
        )



import abc
from dataclasses import asdict, dataclass, field
from typing import Any, Self

from . import Clock
from .data import Metric, MetricChange, MetricChangeRequest


class MessageContent(abc.ABC):
    """Interface of a message content"""

    @property
    def type(self) -> str:
        """Define content type"""
        value = self.__class__.__name__[:-7]
        return value.upper()

    @abc.abstractclassmethod
    def from_dict(cls, value: dict[str, Any]) -> Self:
        """Create content from a dict value"""

    @abc.abstractmethod
    def to_dict(self) -> dict[str, Any]:
        """Convert to plain dict"""


@dataclass
class NbirthContent(MessageContent):
    seq: int
    metrics: list[Metric]

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> Self:
        """Create content from a dict value"""
        metrics = [Metric.from_dict(value_) for value_ in value["metrics"]]
        return cls(value["seq"], metrics)

    def to_dict(self) -> dict[str, Any]:
        """Convert to plain dict"""
        return {
            "metrics": [metric.to_dict() for metric in self.metrics],
            "seq": self.seq,
        }


@dataclass
class NdeathContent(MessageContent):
    bd_seq: Metric

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> Self:
        """Create content from a dict value"""
        return cls(Metric.from_dict(value["metrics"][0]))

    def to_dict(self) -> dict[str, Any]:
        """Convert to plain dict"""
        return {
            "metrics": [self.bd_seq.to_dict()],
        }


@dataclass
class NdataContent(MessageContent):
    seq: int
    changes: list[MetricChange]

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> Self:
        """Create content from a dict value"""
        return cls(
            value["seq"],
            [MetricChange.from_dict(value_) for value_ in value["metrics"]],
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert to plain dict"""
        return {
            "seq": self.seq,
            "metrics": [change.to_dict() for change in self.changes],
        }


@dataclass
class NcmdContent(MessageContent):
    requests: list[MetricChangeRequest]

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> Self:
        """Create content from a dict value"""
        return cls(
            [MetricChangeRequest.from_dict(value_) for value_ in value["metrics"]]
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert to plain dict"""
        return {"metrics": [request.to_dict() for request in self.requests]}


@dataclass
class StateContent(MessageContent):
    online: bool

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> Self:
        """Create content from a dict value"""
        return cls(value["online"])

    def to_dict(self) -> dict[str, Any]:
        """Convert to plain dict"""
        return {"online": self.online}


@dataclass
class WarningContent(MessageContent):
    name: str
    description: str
    args: list = field(default=list)

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> Self:
        """Create content from a dict value"""
        return cls(value["name"], value["description"], value["args"])

    def to_dict(self) -> dict[str, Any]:
        """Convert to plain dict"""
        return asdict(self)


@dataclass(frozen=True)
class Message:
    topic: str
    content: MessageContent
    timestamp: int  # Time in ms when the message was sent

    _CONTENTS = {
        "STATE": StateContent,
        "NBIRTH": NbirthContent,
        "NDATA": NdataContent,
        "NDEATH": NdeathContent,
    }

    @property
    def component_name(self) -> str:
        if "/STATE/" in self.topic:
            return self.topic.split("/")[-1]
        else:
            groups = self.topic.split("/")
            return f"{groups[1]}/{groups[3]}"

    @property
    def content_type(self) -> str:
        if "/STATE/" in self.topic:
            return "STATE"
        else:
            return self.topic.split("/")[2]

    @classmethod
    def create(cls, component_name: str, content: MessageContent, clock: Clock) -> Self:
        if content.type == "STATE":
            topic = f"spBv1.0/STATE/{component_name}"
        else:
            group, node_name = component_name.split("/")
            topic = f"spBv1.0/{group}/{content.type}/{node_name}"
        return cls(topic, content, clock.now())

    @classmethod
    def from_payload(cls, payload: dict[str, Any]) -> Self:
        """Return the topic where the message should be sent"""
        content = payload.copy()
        topic = content.pop("topic")
        if "/STATE/" in topic:
            Content = StateContent
        else:
            Content = cls._CONTENTS[topic.split("/")[2]]

        timestamp = content.pop("timestamp")
        return cls(topic, Content.from_dict(content), timestamp)

    def get_payload(self) -> dict[str, Any]:
        """Return payload to be published"""
        payload = {
            "topic": self.topic,
            "timestamp": self.timestamp,
        }
        payload.update(self.content.to_dict())
        return payload
